##############################################################################
# 
# File            : tsiInit.tcl
# Author          : Gordon Baxter <gdb@vpsyc.psyc.nott.ac.uk>
# Created On      : 16 May 1996, 14:24:30
# 
# PURPOSE
# 
#   This is the initialization interface to the Tcl/TK Soar Interface (TSI),
#   which is designed to work with Soar 7.1 under Unix, Windows, and Macintosh
#
# Copyright 1997, Frank E. Ritter.
###############################################################################
### The Tcl/Tk Soar Interface, for Soar version 7.1
### Major contributors (in alphabetical order):
###    Gordon D. Baxter
###    Randolph M. Jones
###    Douglas J. Pearson
###    Frank E. Ritter
###    Karl B. Schwamb

# safety net in case of typos/misunderstandings
proc TSI {args} {eval tsi $args}

### Call "tsi" to start up the Tcl/Tk Soar Interface.  Pass it an argument
### of 1 if you want to hide the Control Panel (for example, if you just
### want a single Soar agent)

proc tsi {{hideController 0}} {
   global interp_name tsi_version tsi_library tcl_platform tsiConfig 

   ## Make sure tk is loaded, and halt if it is not.

   if {[info commands tk]=={}} {
      puts {You are not running a version of Soar that supports the TSI.
Run a version with TK, contact your system administrator, or email
soar-bugs@umich.edu for help.   

[Load aborted in tsi.tcl]}
      error "Cannot run TSI without the Tk package"
   }

   ### The tsi must be initialized from a non-soar interpreter
   foreach i [info loaded {}] {
      if {[lindex $i 1]=={Soar}} {
         tk_dialog .error Error {The TSI interface can only be initialized from the controlling interpreter.  This should be the top-level interpreter, and it should not have the Soar package loaded.} \
                   error 0 Dismiss
         error "Cannot start TSI from a Soar agent interpreter."
      }
   }

   set tsi_version 2.5

   ### Read in the default TSI parameters
   ### These *must* exist, or the TSI will not run properly
   uplevel #0 [list source [file join $tsi_library tsi-defaults.tcl]]

   ### Read in user parameters if they exist
   
   puts "source'ing user defaults from [pwd]"
   if [file exists [file join [pwd] tsi-defaults.tcl]] {
      uplevel #0 [list source [file join [pwd] tsi-defaults.tcl]]
   }

   tsiInitInterpreter

   makeTSIControlPanel $hideController $tsiConfig(ControlPanelX) \
	                               $tsiConfig(ControlPanelY)

   puts "TSI ver. $tsi_version, expertise: $tsiConfig(expertise), \
debug: $tsiConfig(debug), Platform: $tcl_platform(platform)"

   if {$tsiConfig(debug) && [info exists env(DISPLAY)]} {
      global env
      puts "\nDisplay set to >>$env(DISPLAY)<<"
   }
}

### Call "tsiInitAgent" to initialize a Soar agent interpreter to use
### the TSI
proc tsiInitAgent {} {
   global tsiConfig tsi_library soar7km_aliases
   wm withdraw .

   tsiInitInterpreter

   rename watch tsiInternalWatch
   rename help tsiInternalHelp

   alias p2 print -depth 2
   alias d1 d 1
   alias 1 d 1
   alias 2 d 2
   alias 3 d 3
   alias 4 d 4
   alias 5 d 5
   alias 6 d 6
   alias 7 d 7
   alias 8 d 8
   alias 9 d 9

    if {![info exists soar7km_aliases]} {
	### this should test a global var to see if they are already loaded.
   ## Load the standard aliases
   if [catch {source [file join $tsi_library soar704km-aliases.tcl]} msg] {
      puts "Failed to load Soar KM aliases: $msg"
   } else {
      puts {Soar KM aliases loaded.  Type 'help alias' or 'alias' for more info.}
   }
   }
	set X [expr $tsiConfig(AgentWindowX) + \
			  $tsiConfig(window_count) * $tsiConfig(AgentWindowDeltaX)]
	set Y [expr $tsiConfig(AgentWindowY) + \
			  $tsiConfig(window_count) * $tsiConfig(AgentWindowDeltaY)]

   if [catch {makeSoarInteractionWindow $X $Y} msg] {
      error "Failed to create Soar agent interaction window: $msg"
   }
   monitor -add before-init-soar cleanupMonitorWindows
   ask -add ask-user
}

### Call tsiInitInterpreter to initialize use of the TSI from
### *any* interpreter (including the Control Panel).

### Any time a new interpreter is created, you have to give it the
### correct global auto_path value and then call this procedure to set
### all the other appropriate global variables.
proc tsiInitInterpreter {} {
   ### Do our "nicer" version of the "source" command
   rename source tsiInternalSource
   proc source {args} {
      global tsiSourceArgLast
      if {$args == {}} {
         puts "\[source $tsiSourceArgLast\]"
         ### We have to put the uplevel here because some procedures (like
         ### the package loader) depend on the loaded file sharing a variable
         ### with the calling procedure.  How's that for good programming
         ### style?
         uplevel 1 tsiInternalSource "$tsiSourceArgLast"
      } else {
          set tsiSourceArgLast $args
          uplevel 1 tsiInternalSource $args
      }
   }
}


proc ask-user {options} {

	toplevel .dialog
	set i 0
	foreach o $options {
		tsiOutputToWidget "\nOption $i --> $o"
		print $o
		incr i
	}
	set choice [eval "tk_dialog .dialog Hi \"Choose An Operator\" {} 0 $options"]
	return $choice
}
